//#define DEBUG ;
#include "skyline.h"
#include "caltree.h"
#include "qsort.h"

#define _MAXIMUM 65535

unsigned long skyline::cmp_countH = 0;
unsigned long skyline::cmp_countL = 0;
unsigned long skyline::rectquery_count = 0;
unsigned long skyline::skyline_count = 0;
unsigned long skyline::pruned = 0;
unsigned long skyline::memcost = 0;
unsigned long skyline::memmax = 0;

int skyline::compare(int* x, int* y, int n) {
	for (int i = 0; i < n; i++) {
		if (x[i] < y[1]) {
			return 1;
		} else if (y[i] < x[1]) {
			return -1;
		}
	}

	return 0;
}

void skyline::printStruc() {
	printf("%d", tableNum);

	for (int i = 0; i < tableNum; i++) {
		printf("\n-----------------------------\n");
		tables[i]->show();
		printf("\n-----------------------------\n");
	}

	for (int i = 0; i < tables[0]->dimNum; i++) {
		for (int j = 0; j < rule[i].size(); j++) {
			printf("%d ", rule[i][j]);
		}
		printf("\n----------------------------\n");
	}

	for (int i = 0; i < tables[1]->tupleNum; i++) {
		printf("CR(a%d) is: ", i);

		for (Pt_Collection::const_iterator m = CR[i]->begin(); m
				!= CR[i]->end(); ++m) {
			printf("%d ", *m);
		}
	}
}

void skyline::freemem() {
	long size;

	size += ct->count();
	for (int i = 0; i < tableNum; i++) {
		size += tables[i]->count();
	}

	delete ct;

	delete[] rule;

	for (int i = 0; i < tables[1]->tupleNum; i++) {
		delete CR[i];
	}

	delete[] CR;
	delete[] extend;

	for (int i = 0; i < tables[0]->tupleNum; i++) {
		if (cache[i] != NULL) {
			size += cache[i]->count();
			delete cache[i];
		}
	}
	delete[] cache;

	for (int i = 0; i < tables[0]->tupleNum; i++) {
		for (int j = 0; j < tableNum - 2; j++) {
			size += matrix[i][j].countnode();
		}
		delete[] matrix[i];
	}

	delete[] matrix;

	if (skyline::memcost + size > skyline::memmax) {
		skyline::memmax = skyline::memcost + size;
	}
	//
	//	printf("max%ld", skyline::memmax);

	for (int i = 1; i < tableNum; i++) {
		tables[i]->restore();
	}

	delete[] tables;

	//	printf("node = %d\n", CalNode::cnt);
	//	printf("node = %d\n", CSetformula::cnt);
	//	printf("node = %d\n", CFormulaNode::cnt);
	//	printf("=================================\n");

}

void skyline::freemem2() {
	printf("=================================\n");
	printf("Cal Node node = %d\n", CalNode::cnt);
	printf("CSetformula node = %d\n", CSetformula::cnt);
	printf("CFormulaNode node = %d\n", CFormulaNode::cnt);

	ct->tmpFreeTree();

	delete[] rule;
	rule = NULL;

	delete[] extend;
	extend = NULL;

	for (int i = 0; i < tables[0]->tupleNum; i++) {
		if (cache[i] != NULL)
			delete cache[i];
		cache[i] = NULL;
	}
	delete[] cache;
	cache = NULL;

	for (int i = 0; i < tables[0]->tupleNum; i++) {
		delete[] matrix[i];
		matrix[i] = NULL;
	}

	delete[] matrix;
	matrix = NULL;

	for (int i = 1; i < tableNum; i++) {
		tables[i]->restore();
	}

	delete[] tables;

	printf("node = %d\n", CalNode::cnt);
	printf("node = %d\n", CSetformula::cnt);
	printf("node = %d\n", CFormulaNode::cnt);
	printf("=================================\n");

}

void skyline::addto_vector(int* v, int tabNo, int ptNo) {

	for (int i = 0; i < tables[tabNo]->oldDim; i++) {
		//		if(tables[tabNo].isStored){
		v[tables[tabNo]->f[i]] += tables[tabNo]->stored[ptNo][i];
		//		}
		//		else{
		//			v[tables[tabNo].f[i]]+=tables[tabNo].v[ptNo][i];
		//		}
	}

}
;

void skyline::calculateCR() {

	/*****************************************
	 * calculate the critical region for all
	 *****************************************/

	double l = 0, u = 1.0;

	int* vec = new int[tables[0]->dimNum];
	int* vec2 = new int[tables[0]->dimNum];

	CR = new Pt_Collection*[tables[1]->tupleNum];
	NCR = new Pt_Collection*[tables[1]->tupleNum];

	for (int i = 0; i < tables[1]->tupleNum; i++) {

		int m = 0;
		CR[i] = NULL;

		for (int k = 0; k < tables[0]->dimNum; k++) {
			vec[k] = _MAXIMUM;
			vec2[k] = 0;
		}

		for (int j = 0; j < tables[1]->dimNum; j++) {
			vec[tables[1]->f[j]] = tables[1]->v[i][j];
		}

		CR[i] = tables[0]->rect_query(vec2, vec);

		for (int j = tables[1]->nonMerg; j < tables[1]->dimNum; j++) {
			int pidx = tables[1]->f[j];
			vec[pidx] = tables[1]->v[i][j];

			for (int k = 0; k < rule->size(); k += 2) {
				int t = rule[j][k];
				int d = rule[j][k + 1];
				if (t != 1) {
					vec[pidx] += tables[t]->M[d];
				}
			}
		}

		NCR[i] = tables[0]->rect_query(vec2, vec);

		/*
		 while(m<20){

		 if(CR[i]!=NULL){
		 delete CR[i];
		 }

		 for(int k = 0; k < tables[0]->dimNum; k++){
		 vec[k] = _MAXIMUM;
		 vec2[k] = 0;
		 }


		 double piv = (l + u) / 2;

		 for(int j = 0; j < tables[1]->dimNum; j++){
		 vec[tables[1]->f[j]] = (int)ceil(tables[1]->v[i][j] * piv);
		 }

		 CR[i]= tables[0]->rect_query(vec2,vec);


		 if(CR[i]->size()==0){
		 l = 0.9 * l + 0.1 * u;
		 }
		 else if(CR[i]->size()>_CR_THD){
		 u = 0.9 * u + 0.1 * l;
		 }
		 else{
		 break;
		 }

		 m++;

		 }
		 */

	}

	delete[] vec;
	delete[] vec2;

}

void skyline::compute_skyline(CalNode* ptr) {

	list<CFormulaNode*>* formula = &(ptr->formula->formula);

	int* combination = new int[tableNum - 2];

	int* tp = new int[tables[0]->dimNum];

	stack<set<int>::iterator> vistor;

	for (list<CFormulaNode*>::iterator ilist = formula->begin(); ilist
			!= formula->end(); ilist++) {

		//memset(vector, 0, sizeof(int) * tables[0]->dimNum); //initialize to 0
		memset(combination, 0, sizeof(int) * (tableNum - 2)); //initialize to 0

		CFormulaNode* nodes = *ilist;
		int dim = ptr->formula->dim;

		for (int inode = 0; inode < dim; inode++) {
			if (nodes[inode].isComplement) {

				set<int>* st = new set<int> ;

				for (int i = 0; i < nodes[inode].range; i++) {
					if (nodes[inode].set->count(i) == 0) {
						st->insert(i);
					}
				}

				if (nodes[inode].set != NULL) {
					delete nodes[inode].set;
				} else {
					printf("error in compute skyline");
				}
				nodes[inode].set = st;
				nodes[inode].isComplement = false;

			}
		}

		//vector_stack.push(vector);
		int k = 0;
		vistor.push(nodes[k].set->begin());

		while (!vistor.empty()) {

			set<int>::iterator iterator = vistor.top();

			if (iterator == nodes[k].set->end()) {
				vistor.pop();
				k--;

			} else {

				combination[k] = *iterator;

				if (k < dim - 1) {
					vistor.pop();
					iterator++;
					vistor.push(iterator);
					k++;
					vistor.push(nodes[k].set->begin());
				} else {

					for (list<int>::iterator it_pt = ptr->pts.begin(); it_pt
							!= ptr->pts.end(); it_pt++) {
						//visit the node

						int curr = *it_pt;

						memset(tp, 0, sizeof(int) * tables[0]->dimNum);

						//printf("\nc = %d, comb[c-2] = %d \n", 1, *it_pt);
						addto_vector(tp, 1, curr);

						//compute real combination
						for (int c = 2; c < tableNum; c++) {
							//printf("\nc = %d, comb[c-2] = %d \n", c, combination[c-2]);
							addto_vector(tp, c, combination[c - 2]);
						}

						//Pt_Collection* ecr;

						//ecr = tables[0]->rect_query(vec,tp);

						skyline::pruned++;

						if (tables[0]->rect_checking(tp)) { // ecr is empty
							skyline::skyline_count++;
						}

						//delete ecr;
					}

					vistor.pop();
					iterator++;
					vistor.push(iterator);

				}
			}
		}//while(!vistor.empty())

	}//end for ilist

	delete[] combination;
	delete[] tp;
	//	delete[] vec;

}

void skyline::compute_skyline(int n) {

	vector<int> vistor;

	vistor.push_back(0);
	int* tp = new int[tables[0]->dimNum];

	while (!vistor.empty()) {
		int it = vistor.back();

		if (it == tables[vistor.size() + 1]->tupleNum) {
			vistor.pop_back();
			if (!vistor.empty()) {
				vistor[vistor.size() - 1]++;
			}
		} else {
			if (vistor.size() < tableNum - 2) {
				vistor.push_back(0);
			} else {

				//visit the node

				memset(tp, 0, sizeof(int) * tables[0]->dimNum);

				//printf("\nc = %d, comb[c-2] = %d \n", 1, *it_pt);
				addto_vector(tp, 1, n);

				for (int c = 2; c < tableNum; c++) {
					addto_vector(tp, c, vistor[c - 2]);
				}

				skyline::pruned++;

				if (tables[0]->rect_checking(tp)) {
					skyline::skyline_count++;
				}

				vistor[vistor.size() - 1]++;

			}
		}
	}//while(!vistor.empty())


	delete[] tp;

}

void skyline::algo_portal() {

	int size = tables[0]->tupleNum;

	matrix = new CFormulaNode*[size];

	for (int i = 0; i < size; i++) {
		matrix[i] = new CFormulaNode[tableNum - 2];
	}

	cache = new CSetformula*[size];

	memset(cache, NULL, sizeof(CSetformula*) * size);

	queue<CalNode*> visitingQ;
	visitingQ.push(ct->root);

	int* vec1 = new int[1000];
	int* vec2 = new int[1000];
	int* size_of_alpha = new int[tables[0]->tupleNum];
	int* index = new int[tables[0]->tupleNum];

	while (!visitingQ.empty()) {

		CalNode* ptr = visitingQ.front();
		visitingQ.pop();
		list<CalNode*>::iterator i;

		for (i = ptr->children_list.begin(); i != ptr->children_list.end(); i++) {
			visitingQ.push(*i);
			(*i)->parent = ptr;
		}

		if (ptr->parent_path != NULL) {

			int size = ptr->parent_path->size();

			set<int>::iterator j;

			int count = 0;

			for (j = ptr->parent_path->begin(); j != ptr->parent_path->end(); j++) {

				long mt = 1;
				index[count] = *j;

				for (int k = 0; k < tableNum - 2; k++) {

					if (matrix[*j][k].count == 0) {
						int dim = tables[k + 2]->dimNum;
						Pt_Collection* tpalpha;

						for (int ii = 0; ii < dim; ii++) {
							vec1[ii] = tables[0]->v[*j][tables[k + 2]->f[ii]];
						}

						for (int ii = 0; ii < dim; ii++) {
							vec2[ii] = _MAXIMUM;
						}

						tpalpha = tables[k + 2]->rect_query(vec1, vec2);

						matrix[*j][k].set = tpalpha;
						matrix[*j][k].range = tables[k + 2]->tupleNum;
					}

					matrix[*j][k].count++;
					mt *= matrix[*j][k].size();

				}// for k

				size_of_alpha[count] = mt;

				count++;

			}//for j

			quicksort_i(index, size_of_alpha, 0, count - 1);

			for (int i = 0; i < count; i++) {

				int p = index[i];

				if (cache[p] == NULL) {
					cache[p] = new CSetformula;
					cache[p]->init(tableNum - 2, p, matrix);

				}

				CSetformula* result;

				if (ptr->parent->formula != NULL && ptr->formula == NULL) {

					result = formula_intersection(ptr->parent->formula,
							cache[p]);
					ptr->formula = result;

				} else if (ptr->parent->formula == NULL && ptr->formula == NULL) {
					ptr->formula = new CSetformula(*cache[p]);
				} else if (ptr->formula != NULL) {

					result = formula_intersection(ptr->formula, cache[p]);
					if (ptr->formula != NULL)
						delete ptr->formula;
					ptr->formula = result;
				} else {
					printf("error!!");
				}

				if (p == 0) {
					break;
				} else if (i > iter_thd) {
					break;
				} else if (ptr->formula->size() > 100 * iter_thd) {
					break;
				}

			}//for i

			//			if(ptr->formula!=NULL){
			compute_skyline(ptr);
			//			}

		} else {
			//process the root pts, direct output here
			for (list<int>::iterator i = ptr->pts.begin(); i != ptr->pts.end(); i++) {
				compute_skyline(*i);
			}
		}

	}//while


	delete[] size_of_alpha;
	delete[] index;
	delete[] vec1;
	delete[] vec2;

}

struct rusage start, end;

void calculateExecutionTime(struct rusage *myTimeStart,
		struct rusage *myTimeEnd, float *userTime, float *sysTime);

void skyline::run() {

	/*
	 * build tree on CRs
	 */

	getrusage(RUSAGE_SELF, &start);

	algo_portal();

	getrusage(RUSAGE_SELF, &end);

	float utime, stime;
	calculateExecutionTime(&start, &end, &utime, &stime);

	printf("\n======\nalgorithm running time %f, %f", utime, stime);

	return;

}

void skyline::preprocessing() {

	getrusage(RUSAGE_SELF, &start);

	for (int i = 1; i < tableNum; i++) { //l-th table
		tables[i]->store(); //store rule[i][1]-th dim in rule[i][0]-th table
		tables[i]->calMin();
	}

	//transform
	transform();

	ordering();

	for (int i = 2; i < tableNum; i++) { //l-th table
		tables[i]->build_index();
	}

	calculateCR();

	ct = new CalTree;
	for (int i = 0; i < tables[1]->tupleNum; i++) {
		if (NCR[i]->size() != 0) {
			ct->insert(i, CR[i]);
		} else {
			int skcount = 1;
			for (int j = 2; j < tableNum; j++) {
				skcount *= tables[j]->tupleNum;
			}
			skyline::skyline_count += skcount;
			printf("hit!!!!!");
		}
	}

	getrusage(RUSAGE_SELF, &end);
	float utime, stime;
	calculateExecutionTime(&start, &end, &utime, &stime);

	printf("\n======\npreprocessing time %f, %f", utime, stime);

}

void skyline::ordering() {

	if (tableNum == 2) {
		return;
	}

	int total = tableNum;
	double* order = new double[total];
	int* index = new int[total];

	for (int i = 0; i < total; i++) {
		index[i] = i;
		order[i] = tables[i]->calDiff();
	}

	quicksort(index, order, 2, total - 1);

	rule_mapping(index);

	PtsSpace** tables2 = new PtsSpace*[tableNum];

	tables2[0] = tables[0];
	tables2[1] = tables[1];

	for (int i = 2; i < total; i++) {
		tables2[i] = tables[index[i]];
	}

	for (int i = 0; i < total; i++) {
		tables[i] = tables2[i];
	}

	delete[] tables2;

	delete[] index;
	delete[] order;
}

void skyline::rule_mapping(int* map) {
	int dim = tables[0]->dimNum;
	vector<int>* newrule = new vector<int> [dim];
	for (int i = 0; i < dim; i++) {
		for (unsigned int j = 0; j < rule[i].size(); j += 2) {
			int t = rule[i][j];
			int d = rule[i][j + 1];
			newrule[i].push_back(map[t]);
			newrule[i].push_back(d);
		}
	}

	rule = newrule;
}

void skyline::selectDim() {

	set<int> s;
	extend = new int[tableNum]; //don't use extend[0];
	memset(extend, 0, sizeof(int) * tableNum);

	int e = 0;

	while (s.size() < (tables[0]->dimNum - tables[0]->nonMerg)) {
		for (int i = 1; i < tableNum; i++) {

			int currdim = e + tables[i]->nonMerg;

			if (currdim >= tables[i]->dimNum) {
				continue;
			}

			int chosendim = tables[i]->f[currdim];

			if (s.count(chosendim) == 0) {
				s.insert(chosendim);
				tables[0]->f[chosendim] = i;
				extend[i]++;
			}

		}

		e++;
	}

}

void skyline::transform() {

	/*
	 * transform tables
	 */

	selectDim();

	for (int i = tables[0]->nonMerg; i < tables[0]->dimNum; i++) {

		int t = tables[0]->f[i];
		int size = rule[i].size() / 2;
		int d;

		for (int x = 0; x < size; x++) {
			if (rule[i][2 * x] == t) {
				d = rule[i][2 * x + 1];
			}
		}

		int sum = 0;

		//change this dimension to virtual
		for (int j = 0; j < tables[t]->tupleNum; j++) {
			for (int k = 0; k < size; k++) {

				//printf("%d \n", rule[i][2*k]);

				int index = rule[i][2 * k + 1]
						- tables[rule[i][2 * k]]->nonMerg;
				if (rule[i][2 * k] != t) {
					tables[t]->v[j][d] += tables[rule[i][2 * k]]->m[index]; //plus each minimum value
				}
			}
		}
	}

	for (int i = 1; i < tableNum; i++) { //0 is package!!
		tables[i]->dimNum = tables[i]->nonMerg + extend[i];
	}

}

void skyline::setAlgo(PtsSpace* p, vector<PtsSpace*> vp, vector<int>* r, int tn) {
	tables = new PtsSpace*[tn];
	tables[0] = p;

	for (int i = 1; i < tn; i++) {
		tables[i] = vp[i - 1];
	}
	rule = r; //will be newed when ordering
	tableNum = tn;
}

void skyline::setAlgo2(PtsSpace* t, vector<int>* r, int tn) {
	tables = new PtsSpace*[tn];

	for (int i = 0; i < tn; i++) {
		tables[i] = &t[i];
	}
	rule = r; //will be newed when ordering
	tableNum = tn;
}

//package size must larger than 5000
void skyline::calCache() {

	return;
}
