/*
 * space.cpp
 *
 *  Created on: Jan 14, 2009
 *      Author: qwan
 */

#include "space.h"
#include "skyline.h"
#include "qsort.h"
#include <time.h>

//#define SORTPOS 2

using namespace std;

CSetformula* formula_intersection(CSetformula* f1, CSetformula* f2);
CFormulaNode* node_intersection(CFormulaNode* set1, CFormulaNode* set2);

unsigned long PtsSpace::totalAsked = 0;
unsigned long PtsSpace::cacheHits = 0;
unsigned long PtsSpace::cacheMissed = 0;
int PtsSpace::cnt = 0;

void initialize(config_type *config, char *configFile, rtree_info *aInfo);
void
		build_tree(node_type **root, R_TYPE **data, int no_data,
				rtree_info *aInfo);

void free_tree(node_type *node, rtree_info *aInfo);

void PtsSpace::printVec(int* v, int n) {
	for (int i = 0; i < n; i++) {
		printf("%d ", v[i]);
	}
}
;

int PtsSpace::dominate(const int* v1, const int* v2) {
	{
		bool v1grt = false, v2grt = false;
		for (int i = 0; i < dimNum; i++) {
			if (v1[i] < v2[i]) {
				v1grt = true;
			} else if (v2[i] < v1[i]) {
				v2grt = true;
			}

			if (v1grt && v2grt) {
				return 2; //no dominate relation
			}

		}

		if (v1grt) {
			return 1; //v1 dominate v2
		} else if (v2grt) {
			return -1; //v2 dominate v1
		} else {
			return 0; //v1 = v2
		}
	}
}

int PtsSpace::dominate(const int* v1, const int* v2, int n) {
	bool v1grt = false, v2grt = false;

	for (int i = 0; i < n; i++) {
		if (v1[i] < v2[i]) {
			v1grt = true;
		} else if (v2[i] < v1[i]) {
			v2grt = true;
		}
		if (v1grt && v2grt) {
			return 2; //no dominate relation
		}
	}

	if (v1grt) {
		return 1; //v2 dominate v1
	} else if (v2grt) {
		return -1; //v1 dominate v2
	} else {
		return 0; //v1 = v2
	}
}
;

void PtsSpace::store() {

	if (isStored) {
		return;
	} else {
		for (int i = 0; i < tupleNum; i++) {
			memcpy(stored[i], v[i], sizeof(int) * dimNum);
		}

		isStored = true;

	}

}
;

void PtsSpace::restore() {

	if (!isStored)
		return;

	for (int i = 0; i < tupleNum; i++) {
		memcpy(v[i] + nonMerg, stored[i] + nonMerg, sizeof(int) * (dimNum
				- nonMerg));

	}

	//	delete[] v;
	dimNum = oldDim;
	//	v = stored;

	//	stored = new int*[tupleNum];
	//
	//	for(int i=0;i<tupleNum;i++){
	//	  stored[i] = new int[dimNum];
	//	}

	//isStored = false;

	//	release_index();

}

Pt_Collection* PtsSpace::rect_query(const int* v1, const int* v2) {

	Pt_Collection* result;

	result = new Pt_Collection;

	int re1 = 0; //result compare with v1;
	int re2 = 0; //result compare with v2;

	if (indexed_flag) {
		rectangle_search(r_root, v1, v2, result, false);
	} else {
		for (int i = 0; i < tupleNum; i++) { //for each tuple
			re1 = dominate(v1, v[i]);
			re2 = dominate(v[i], v2);
			if ((re1 == 1 || re1 == 0) && (re2 == 1 || re2 == 0)) {
				result->insert(i);
			}
		}

	}

	return result;

}

void PtsSpace::rectangle_search(node_type *curr_node, const int *u,
		const int* l, Pt_Collection* result, bool inside) {
	int i, stop;
	int query_dim;

	query_dim = r_aInfo.dim;
	stop = r_aInfo.M - curr_node->vacancy;

	if (inside) {
		if (curr_node->ptr[0]->attribute == LEAF) {
			for (i = 0; i < stop; i++) {
				result->insert(curr_node->ptr[i]->id);
			}
			return;
		}

		for (i = 0; i < stop; i++) {
			/* search the node which contains the query */
			rectangle_search(curr_node->ptr[i], u, l, result, inside);
		}

	} else {
		if (curr_node->ptr[0]->attribute == LEAF) {
			for (i = 0; i < stop; i++) {
				int cmp_l = dominate(curr_node->ptr[i]->a, l, query_dim);
				int cmp_u = dominate(u, curr_node->ptr[i]->a, query_dim);
				if ((cmp_l == 1 || cmp_l == 0) && (cmp_u == 1 || cmp_u == 0)) {
					result->insert(curr_node->ptr[i]->id);
				}
			}

			return;
		}

		for (i = 0; i < stop; i++) {

			/*
			 * a must equal or dominate l
			 * u must equal dominate b
			 *
			 * inside: a equals or dominates u && l equals or dominates b
			 * otherwise outside
			 *
			 * */

			int a_l = dominate(curr_node->ptr[i]->a, l, query_dim);

			if (a_l == -1 || a_l == 2) {
				continue;
			} else {
				int b_u = dominate(curr_node->ptr[i]->b, u, query_dim);

				if (b_u == 1 || b_u == 2) {
					continue;
				}

				int b_l = dominate(curr_node->ptr[i]->b, l, query_dim);

				if ((b_l == 0 || b_l == 1)) {
					int u_a = dominate(u, curr_node->ptr[i]->a, query_dim);
					if (u_a == 0 || u_a == 1) {
						rectangle_search(curr_node->ptr[i], u, l, result, true);
					} else {
						rectangle_search(curr_node->ptr[i], u, l, result, false);
					}
				} else {
					rectangle_search(curr_node->ptr[i], u, l, result, false);
				}
			}

		}
		//
		//			if(a_l == 1 || a_l == 0){
		//				int u_b = dominate(u,curr_node->ptr[i]->b, query_dim);
		//				if(u_b == 1 || u_b == 0){
		//					int a_u = dominate(curr_node->ptr[i]->a,u,query_dim);
		//					if(a_u == 1 || a_u == 0){
		//						int l_b = dominate(l,curr_node->ptr[i]->b, query_dim);
		//							if(l_b == 1 || l_b == 0){
		//								rectangle_search(curr_node->ptr[i], u, l, result, true);
		//							}
		//							else
		//							{
		//								rectangle_search(curr_node->ptr[i], u, l, result, inside);
		//							}
		//					}
		//					else
		//					{
		//						rectangle_search(curr_node->ptr[i], u, l, result, inside);
		//					}//end of
		//				}//endif of if(u_b == 1 || u_b == 0)
		//			}//endif of if(a_l == 1 || a_l == 0)
		//		}//end of for
	}

	return;

}

void PtsSpace::build_index() {

	if (indexed_flag)
		return;

	//initialize(&config, "./r_config", &r_aInfo);

	r_aInfo.m = 2 * (int) pow(tupleNum, 1.0 / 5) - 1;

	if (r_aInfo.m < 3) {
		r_aInfo.m = 3;
	}

	r_aInfo.M = r_aInfo.m * 2;
	r_aInfo.reinsert_p = r_aInfo.m;

	printf("rtree building, tuple = %d, M = %d, m = %d, reinsert = %d\n",
			tupleNum, r_aInfo.M, r_aInfo.m, r_aInfo.reinsert_p);

	r_aInfo.dim = this->dimNum;
	r_aInfo.no_histogram = this->tupleNum;
	//int** data;
	if (isPackage) {
		r_data = v;
	} else {
		r_data = v;
	}

	PtsSpace::cnt++;

	build_tree(&r_root, r_data, tupleNum, &r_aInfo);

	indexed_flag = true;

}

void PtsSpace::release_index() {

	if (indexed_flag) {

		PtsSpace::cnt--;

		indexed_flag = false;
		free_tree(r_root, &r_aInfo);
		r_root = NULL;
	}

}

//filter for unskyline points
void PtsSpace::bnl() {

	sort(nonMerg);

	int length = 0;

	v2 = new int*[tupleNum];

	for (int j = 0; j < tupleNum; j++) {
		if (length == 0) {
			v2[length] = new int[dimNum];
			memcpy(v2[length], v[j], sizeof(int) * dimNum);
			length++;
		} else {
			addpoint(&length, j);
			while (mem.size() > 1000000) {
				delete mem.top();
				mem.pop();
			}
		}
	}

	for (int i = 0; i < tupleNum; i++) {
		delete[] v[i];
	}

	delete[] v;

	while (mem.size() > 0) {
		delete mem.top();
		mem.pop();
	}

	tupleNum = length;

	v = v2;

}
;

int PtsSpace::dnlcomp(int* vec1, int* vec2) {

	for (int i = sortpos - 1; i >= 0; i--) {
		if (vec1[i] > vec2[i])
			return -1;
	}

	for (int i = sortpos + 1; i < dimNum; i++) {
		if (vec1[i] > vec2[i])
			return -1;
	}

	return 1; //v1 = v2
}

void PtsSpace::addpoint(int* len, int ptr) {
	int postion = *len - 1;
	bool insertNeed = true;

	while (postion >= 0 && v2[postion][sortpos] == v[ptr][sortpos]) {
		int re = dominate(v[ptr], v2[postion]);
		if (re == 1 || re == 0) {
			//delete[] v2[postion];
			mem.push(v2[postion]);
			v2[postion] = NULL;
		} else if (re == -1) {
			insertNeed = false;
		}
		postion--;
	}

	if (insertNeed) {
		for (int i = postion; i >= 0; i--) {
			if (dnlcomp(v[ptr], v2[i]) == 1) {
				//delete[] v2[i];
				mem.push(v2[i]);
				v2[i] = NULL;
			}
		}
	}

	int count = 0;

	for (int i = 0; i < *len; i++) {
		if (v2[i] != NULL) {
			v2[count] = v2[i];
			count++;
		}
	}

	if (insertNeed) {

		if (mem.size() != 0) {
			v2[count] = mem.top();
			mem.pop();
		} else {
			v2[count] = new int[dimNum];
		}

		memcpy(v2[count], v[ptr], sizeof(int) * dimNum);

		*len = count + 1;
	}

	return;
}

void PtsSpace::getdata(FILE *fp) {
	for (int i = 0; i < tupleNum; i++) {
		for (int j = 0; j < dimNum; j++) {
			fscanf(fp, "%d", &v[i][j]);
		}
	}

	//fclose(fp);
}

void PtsSpace::calMin() {
	for (int i = nonMerg; i < dimNum; i++) {
		m[i - nonMerg] = v[0][i];
		for (int j = 1; j < tupleNum; j++) {
			if (m[i - nonMerg] > v[j][i]) {
				m[i - nonMerg] = v[j][i];
			}
		}
	}
}

void PtsSpace::swap(int d1, int d2) {
	int temp;

	if (d1 < nonMerg || d2 < nonMerg) {
		printf("error");
		return;
	}

	for (int i = 0; i < tupleNum; i++) {
		temp = v[i][d1];
		v[i][d1] = v[i][d2];
		v[i][d2] = temp;
	}
}

double PtsSpace::sp_dist(int* v1, int* v2) {
	double sum = 0;

	for (int i = nonMerg; i < dimNum; i++) {
		sum = sum + (double) (v1[i] - v2[i]) * (v1[i] - v2[i]);
	}

	return sqrt(sum);
}

double PtsSpace::sp_dist2(int* v1, int* v2) {
	double sum = 0;
	for (int i = 0; i <= nonMerg; i++) {
		sum = sum + (double) (v1[i] - v2[i]) * (v1[i] - v2[i]);
	}
	return sqrt(sum);
}

void addvec(int**v, int**v2, int dim) {
	for (int i = 0; i < dim; i++) {
		(*v)[i] += (*v2)[i];
	}
}

Pt_Collection* PtsSpace::KMeans2(int n) {
	srand(time(0));
	Pt_Collection* p = new Pt_Collection[n];
	Pt_Collection marked;

	int recal_threhold = 1;

	int** matrix = new int*[n];

	for (int i = 0; i < n; i++) {
		matrix[i] = new int[dimNum];
		int idx = tupleNum / n * i;
		memcpy(matrix[i], v[idx], sizeof(int) * dimNum);
		p[i].insert(idx);
		marked.insert(idx);
	}

	for (int m = 0; m < 1; m++) {

		if (m > 0) {
			marked.clear();

			for (int i = 0; i < n; i++) {
				int x = *p[i].rbegin();
				p[i].clear();
				p[i].insert(x);
				marked.insert(x);
			}
		}

		for (int i = 0; i < tupleNum; i++) {

			if (marked.count(i) != 0) {
				continue;
			}

			double min_dist = 90000000.0;
			int min_mean = 0;

			for (int j = 0; j < n; j++) {
				double re = sp_dist2(v[i], matrix[j]);
				if (re < min_dist) {
					min_dist = re;
					min_mean = j;
				}
			}
			p[min_mean].insert(i);

			if (i % recal_threhold == 0) {

				for (int j = 0; j < n; j++) {

					Pt_Collection::iterator iter;
					int* avg = new int[dimNum];
					memset(avg, 0, sizeof(int) * dimNum);
					for (iter = p[j].begin(); iter != p[j].end(); iter++) {
						addvec(&avg, &(v[*iter]), dimNum);
					}
					for (int k = 0; k < dimNum; k++) {
						avg[k] /= p[j].size();
					}

					memcpy(matrix[j], avg, sizeof(int) * dimNum);
					delete[] avg;
				}
			}

		}
	}

	for (int i = 0; i < n; i++) {
		delete[] matrix[i];
	}

	delete[] matrix;
	return p;
}

Pt_Collection* PtsSpace::KMeans(int n) {
	srand(time(0));
	Pt_Collection* p = new Pt_Collection[n];
	Pt_Collection marked;

	int recal_threhold = 1;

	int** matrix = new int*[n];

	for (int i = 0; i < n; i++) {
		matrix[i] = new int[dimNum];
		int idx = tupleNum / n * i;
		memcpy(matrix[i], v[idx], sizeof(int) * dimNum);
		p[i].insert(idx);
		marked.insert(idx);
	}

	for (int m = 0; m < 1; m++) {

		if (m > 0) {
			marked.clear();

			for (int i = 0; i < n; i++) {
				int x = *p[i].rbegin();
				p[i].clear();
				p[i].insert(x);
				marked.insert(x);
			}
		}

		for (int i = 0; i < tupleNum; i++) {

			if (marked.count(i) != 0) {
				continue;
			}

			double min_dist = 90000000.0;
			int min_mean = 0;

			for (int j = 0; j < n; j++) {
				double re = sp_dist(v[i], matrix[j]);
				if (re < min_dist) {
					min_dist = re;
					min_mean = j;
				}
			}
			p[min_mean].insert(i);

			if (i % recal_threhold == 0) {

				for (int j = 0; j < n; j++) {

					Pt_Collection::iterator iter;
					int* avg = new int[dimNum];
					memset(avg, 0, sizeof(int) * dimNum);
					for (iter = p[j].begin(); iter != p[j].end(); iter++) {
						addvec(&avg, &(v[*iter]), dimNum);
					}
					for (int k = 0; k < dimNum; k++) {
						avg[k] /= p[j].size();
					}

					memcpy(matrix[j], avg, sizeof(int) * dimNum);
					delete[] avg;
				}
			}

		}
	}

	for (int i = 0; i < n; i++) {
		delete[] matrix[i];
	}

	delete[] matrix;
	return p;
}

//PtsSpace* PtsSpace::divid(){
//	int num = 1<<nonMerg;
//	int* med = new int[nonMerg];
//	PtsSpace* array = ;
//
//
//}

PtsSpace* PtsSpace::split(int n) {
	int m = 1;

	Pt_Collection* p = KMeans(n);

	PtsSpace* pArray = new PtsSpace[n * m];

	for (int i = 0; i < n; i++) {
		PtsSpace* pts = Product(&p[i]);
		pts->sort(0);
		Pt_Collection* p2 = pts->KMeans(m);
		for (int j = 0; j < m; j++) {
			PtsSpace* pts2 = Product(&p2[j]);
			pArray[i * n + j] = *pts2;
			delete pts2;
		}

		delete pts;
	}

	delete[] p;
	return pArray;
}

PtsSpace* PtsSpace::split(int n, int m, int* g) {
	int count = 0;
	Pt_Collection* p = KMeans(n);

	PtsSpace* pArray = new PtsSpace[n * m];

	for (int i = 0; i < n; i++) {
		PtsSpace* pts = Product(&p[i]);

		if (pts->tupleNum > 2 * m) {
			pts->sort(nonMerg - 1);
			Pt_Collection* p2 = pts->KMeans2(m);
			for (int j = 0; j < m; j++) {
				PtsSpace* pts2 = pts->Product(&p2[j]);
				//pts2->calM();
				pArray[count].init(pts2->dimNum, pts2->tupleNum, pts2->nonMerg);
				for (int i = 0; i < pts2->dimNum; i++) {
					for (int j = 0; j < pts2->tupleNum; j++) {
						pArray[count].v[j][i] = pts2->v[j][i];
					}

					pArray[count].f[i] = pts2->f[i];

				}
				count++;
				delete pts2;
			}

			delete[] p2;
		} else {
			pArray[count].init(pts->dimNum, pts->tupleNum, pts->nonMerg);
			for (int i = 0; i < pts->dimNum; i++) {
				for (int j = 0; j < pts->tupleNum; j++) {
					pArray[count].v[j][i] = pts->v[j][i];
				}

				pArray[count].f[i] = pts->f[i];

			}
			count++;
		}

		delete pts;
	}

	//	for(int i=0;i<count;i++)
	//		pArray[i].show();

	*g = count;

	delete[] p;
	return pArray;
}

void PtsSpace::sort(int d) {

	sortpos = d;

	int* temp = new int[tupleNum];
	int* idx = new int[tupleNum];
	for (int i = 0; i < tupleNum; i++) {
		temp[i] = v[i][d];
		idx[i] = i;
	}

	quicksort_i(idx, temp, 0, tupleNum - 1);

	int** v2 = new int*[tupleNum];

	for (int i = 0; i < tupleNum; i++) {
		v2[i] = v[idx[i]];
	}

	delete[] v;
	delete[] temp;
	delete[] idx;
	v = v2;

}

void PtsSpace::addto_dim(int d, int mount) {
}

//calculate the rate for ordering
double PtsSpace::calDiff() {
	int* max = new int[dimNum - nonMerg];
	int* min = new int[dimNum - nonMerg];
	double result = 0;

	for (int j = 0; j < dimNum - nonMerg; j++) {
		max[j] = 0;
		min[j] = _MAXIMUM;
	}

	for (int i = 0; i < tupleNum; i++) {
		for (int j = 0; j < dimNum - nonMerg; j++) {
			if (v[i][nonMerg + j] > max[j]) {
				max[j] = v[i][nonMerg + j];
			}

			if (v[i][nonMerg + j] < min[j]) {
				min[j] = v[i][nonMerg + j];
			}
		}
	}

	for (int j = 0; j < dimNum - nonMerg; j++) {
		result += 1.0 / (max[j] - min[j]);
	}

	delete[] max;
	delete[] min;

	return 1.0 / result;
}

double PtsSpace::calDiff2() {

	int sum1 = 0, sum2 = 0;

	for (int i = 0; i < tupleNum; i += 3) {
		int k = rand() % tupleNum;
		sum1 += v[k][nonMerg];
		sum2 += v[k][nonMerg] * v[k][nonMerg];
	}

	double factor = tupleNum / 3.0;
	double avg1 = (sum1 / factor) * (sum1 / factor);
	double avg2 = sum2 / factor;

	return sqrt(avg2 - avg1);
}

void PtsSpace::calM() {

	for (int i = 0; i < dimNum; i++) {
		min[i] = M[i] = v[0][i];
		for (int j = 1; j < tupleNum; j++) {
			if (M[i] < v[j][i]) {
				M[i] = v[j][i];
			}

			if (min[i] > v[j][i]) {
				min[i] = v[j][i];
			}

		}
	}

}

//void calM2(){
//
//	for(int i=0;i<dimNum;i++){
//		min[i] = M[i] = v[0][i];
//		for(int j=1;j<tupleNum;j++){
//			if(M[i]<v[j][i]){
//				M[i] = v[j][i];
//			}
//
//			if(min[i]>v[j][i]){
//				min[i] = v[j][i];
//			}
//
//		}
//	}
//
//}

PtsSpace* PtsSpace::Product(Pt_Collection* p) {

	PtsSpace* result = new PtsSpace;
	int size = p->size();

	result->init(dimNum, size, nonMerg);

	if (result->tupleNum == 0) {
		return result;
	}

	int count = 0;
	for (Pt_Collection::iterator i = p->begin(); i != p->end(); i++) {
		int index = *i;
		memcpy(result->v[count++], v[index], sizeof(int) * dimNum);
	}

	memcpy(result->f, f, sizeof(int) * dimNum);

	return result;

}

PtsSpace* PtsSpace::ProductOrg(int index) {

	PtsSpace* result = new PtsSpace;
	result->init(oldDim, 1, nonMerg);

	memcpy(result->v[0], stored[index], sizeof(int) * oldDim);

	memcpy(result->f, f, sizeof(int) * oldDim);

	return result;

}

bool PtsSpace::rect_checking(int* q) {

	PtsSpace::totalAsked++;
	int* zero = new int[dimNum];
	memset(zero, 0, sizeof(int) * dimNum);
	bool re = true;

	if (indexed_flag) {
		re = !rectangle_find(r_root, zero, q);
	} else {
		for (int i = 0; i < tupleNum; i++) { //for each tuple
			//int re1 = dominate(zero, v[i]);
			int re2 = dominate(v[i], q);
			if (re2 == 1 || re2 == 0) {
				re = false;
				break;
			}
		}
	}
	delete[] zero;

	return re;

	/*
	 PtsSpace::totalAsked++;
	 int* zero =new int[dimNum];
	 memset(zero,0,sizeof(int)*dimNum);

	 if(PC != NULL){
	 //printf("try PC...");
	 Pt_Collection* cr = PC->rect_query(zero,q);
	 if(cr->size() != 0){
	 //printf("match once!!!");
	 PtsSpace::cacheHits++;
	 delete zero;
	 delete cr;
	 return false;
	 }
	 delete cr;
	 }

	 Pt_Collection* cr = rect_query(zero,q);

	 if(cr->size()==0){
	 delete cr;
	 delete zero;
	 return true;
	 }
	 else
	 {
	 PtsSpace::cacheMissed++;
	 delete cr;
	 delete zero;
	 return false;
	 }
	 */
}

PtsSpace* PtsSpace::getSample() {

	int size = tupleNum;

	if (size < 100) {
		return NULL;
	}

	int* q = new int[size];
	int* zero = new int[size];
	memset(zero, 0, sizeof(int) * size);

	calM();

	Pt_Collection* re = NULL;
	double l = 0, u = 1.0;
	int k = 0;
	int thd = (int) sqrt(tupleNum);
	//int lg = ;

	while (k < 50) {

		k++;
		double piv = (u + l) / 2;

		for (int i = 0; i < dimNum; i++) {
			q[i] = (int) floor(M[i] * piv);
		}

		if (re != NULL) {
			delete re;
		}

		re = rect_query(zero, q);

		int res = re->size();

		if (res < tupleNum / 9 * thd) {
			l = 0.99 * l + 0.01 * u;
		} else if (res > tupleNum / 10 * thd) {
			u = 0.99 * u + 0.01 * l;
		} else {
			break;
		}

	}

	if (re->size() >= 9 * thd) {
		printf("use a %d size cache instead of %d size package\n", re->size(),
				tupleNum);
		PC = Product(re);
		PC->build_index();
	}

	delete[] q;
	delete[] zero;
	delete re;

}

void PtsSpace::sampling() {

	if (sample != NULL) {
		printf("use a %d size cache instead of %d size package\n",
				sample->size(), tupleNum);
		if (PC != NULL) {
			delete PC;
		}
		PC = Product(sample);
		return;
	}

	int size = tupleNum;

	//	if(size < 100){
	//		return;
	//	}

	int* q = new int[size];
	int* zero = new int[size];
	memset(zero, 0, sizeof(int) * size);

	calM();

	Pt_Collection* re = NULL;
	double l = 0, u = 1.0;
	int k = 0;
	//int thd = (int)sqrt(tupleNum)/2;
	//int lg = ;

	while (k < 50) {

		k++;
		double piv = (u + l);

		for (int i = 0; i < dimNum; i++) {
			if (i < nonMerg) {
				q[i] = (int) floor(M[i] * piv * 0.9);
			} else {
				q[i] = (int) floor(M[i] * piv);
			}
		}

		if (re != NULL) {
			delete re;
		}

		re = rect_query(zero, q);

		int res = re->size();

		if (res < tupleNum / 101 * 99) {
			l = 0.7 * l + 0.3 * u;
		} else if (res > tupleNum / 101 * 100) {
			u = 0.7 * u + 0.3 * l;
		} else {
			break;
		}

	}

	if (re->size() != 0) {
		printf("use a %d size cache instead of %d size package\n", re->size(),
				tupleNum);
		if (PC != NULL) {
			delete PC;
		}
		PC = Product(re);
		sample = re;
		//PC->build_index();
	}

	delete[] q;
	delete[] zero;
	//delete re;

}

void PtsSpace::remove(Pt_Collection& s) {
	for (int i = 0; i < tupleNum; i++) {
		if (s.count(i) == 1) {
			delete v[i];
			v[i] = NULL;
		}
	}
	int count = 0;

	for (int i = 0; i < tupleNum; i++) {
		if (v[i] == NULL) {
			continue;
		}
		v[count++] = v[i];
	}

	tupleNum = count;

}

void PtsSpace::prunning() {
}
;

PtsSpace* PtsSpace::Selection(int* map, int dim, int non) {

	PtsSpace* result = new PtsSpace;
	result->init(dim, tupleNum, non);

	for (int i = 0; i < tupleNum; i++) {
		for (int j = 0; j < dim; j++) {
			result->v[i][j] = v[i][map[j]];
		}
	}

	return result;

}

bool PtsSpace::rectangle_find(node_type *curr_node, const int *u, const int* l) {

	int i, stop;
	int query_dim;

	query_dim = r_aInfo.dim;
	stop = r_aInfo.M - curr_node->vacancy;

	if (curr_node->ptr[0]->attribute == LEAF) {
		for (i = 0; i < stop; i++) {
			int cmp_l = dominate(curr_node->ptr[i]->a, l, query_dim);
			int cmp_u = dominate(u, curr_node->ptr[i]->a, query_dim);
			if ((cmp_l == 1 || cmp_l == 0) && (cmp_u == 1 || cmp_u == 0)) {
				return true;
			}
		}
		return false;
	}

	for (i = 0; i < stop; i++) {

		/*
		 * a must equal or dominate l
		 * u must equal dominate b
		 *
		 * inside: a equals or dominates u && l equals or dominates b
		 * otherwise outside
		 *
		 * */

		int a_l = dominate(curr_node->ptr[i]->a, l, query_dim);

		if (a_l == -1 || a_l == 2) {
			continue;
		} else {
			int b_u = dominate(curr_node->ptr[i]->b, u, query_dim);

			if (b_u == 1 || b_u == 2) {
				continue;
			}

			int b_l = dominate(curr_node->ptr[i]->b, l, query_dim);

			if (b_l == 0 || b_l == 1) {
				int u_a = dominate(u, curr_node->ptr[i]->a, query_dim);
				if (u_a == 0 || u_a == 1) {
					return true;
				} else {
					if (rectangle_find(curr_node->ptr[i], u, l)) {
						return true;
					}
				}
			} else {
				if (rectangle_find(curr_node->ptr[i], u, l)) {
					return true;
				}
			}
		}

	}

	return false;

	//		int a_l = dominate(curr_node->ptr[i]->a, l, query_dim);
	//
	//		if(a_l == 1 || a_l == 0){
	//			int u_b = dominate(u,curr_node->ptr[i]->b, query_dim);
	//			if(u_b == 1 || u_b == 0){
	//				int a_u = dominate(curr_node->ptr[i]->a,u,query_dim);
	//				if(a_u == 1 || a_u == 0){
	//					int l_b = dominate(l,curr_node->ptr[i]->b, query_dim);
	//						if(l_b == 1 || l_b == 0){
	//							return true;
	//						}
	//						else
	//						{
	//							if(rectangle_find(curr_node->ptr[i], u, l)){
	//								return true;
	//							}
	//						}
	//				}
	//				else
	//				{
	//					if(rectangle_find(curr_node->ptr[i], u, l)){
	//						return true;
	//					}				}//end of
	//			}//endif of if(u_b == 1 || u_b == 0)
	//		}//endif of if(a_l == 1 || a_l == 0)	int i, stop;
	//	}

}

void PtsSpace::minusMin() {
	for (int i = 0; i < tupleNum; i++) {
		for (int j = nonMerg; j < dimNum; j++) {
			v[i][j] -= min[j];
		}
	}
}
